"""
Functions for the API.

.. _packages: https://pypi.org/search/?q=%22Measurement+Standards+Laboratory+of+New+Zealand%22
.. _git: https://git-scm.com
.. _repositories: https://github.com/MSLNZ
.. _JSON: https://www.json.org/
"""
import base64
import collections
import datetime
import fnmatch
import getpass
import json
import logging
import os
import platform
import re
import shlex
import struct
import subprocess
import sys
import textwrap
import threading
import time

try:
    from importlib import reload
    from urllib.request import urlopen, Request, HTTPError, URLError
except ImportError:  # then Python 2
    from imp import reload
    from urllib2 import urlopen, Request, HTTPError, URLError

import pkg_resources
from colorama import Back
from colorama import Fore
from colorama import Style
from colorama import init
from pkg_resources import packaging

_PKG_NAME = 'msl-package-manager'
_pip_quiet = 0
_IS_WINDOWS = sys.platform == 'win32'

_HOME_DIR = os.path.join(os.path.expanduser('~'), '.msl', 'package-manager')
if not os.path.isdir(_HOME_DIR):
    os.makedirs(_HOME_DIR)

_GITHUB_AUTH_PATH = os.path.join(_HOME_DIR, 'github-auth')

# the HOME_DIR changed from ~/.msl to ~/.msl/package-manager
# move a previously-created github-auth file to the new HOME_DIR
_old_auth_path = os.path.join(os.path.expanduser('~'), '.msl', '.mslpm-github-auth')
if os.path.isfile(_old_auth_path):
    os.rename(_old_auth_path, _GITHUB_AUTH_PATH)

_package_name_regex = re.compile(
    r'(?P<package_name>[*]?[\w-]*[*]?[\w-]*)(?P<extras_require>\[.*\])?(?P<version_requested>[<!=>~].*)?'
)

try:
    subprocess.check_output(['git', '--version'])
except:
    has_git = False
else:
    has_git = True


def get_email():
    """Try to determine the user's email address.

    If git_ is installed then it returns the ``user.email`` parameter from the user's git_
    account to use as the user's email address. If git_ is not installed then returns
    :data:`None`.

    Returns
    -------
    :class:`str` or :data:`None`
        The user's email address.
    """
    if has_git:
        try:
            email = subprocess.check_output(['git', 'config', 'user.email'])
        except subprocess.CalledProcessError:
            pass
        else:
            return email.strip().decode('utf-8')


def get_username():
    """Determine the name of the user.

    If git_ is installed then it returns the ``user.name`` parameter from the user's git_
    account. If git_ is not installed or if the ``user.name`` parameter does not exist
    then :func:`getpass.getuser` is used to determine the username.

    Returns
    -------
    :class:`str`
        The user's name.
    """
    if has_git:
        try:
            uname = subprocess.check_output(['git', 'config', 'user.name'])
        except subprocess.CalledProcessError:
            pass
        else:
            return uname.strip().decode('utf-8')
    return getpass.getuser()


def github(update_cache=False):
    """Get the information about the MSL repositories_ that are available on GitHub.

    Parameters
    ----------
    update_cache : :class:`bool`, optional
        The information about the repositories_ that are available on GitHub are
        cached to use for subsequent calls to this function. After 24 hours the
        cache is automatically updated. Set `update_cache` to be :data:`True`
        to force the cache to be updated when you call this function.

    Returns
    -------
    :class:`dict`
        The information about the MSL repositories_ that are available on GitHub.
    """
    packages, path = _inspect_github_pypi('github', update_cache)
    if packages:
        return packages

    def fetch(url_suffix):
        url = 'https://api.github.com' + url_suffix
        try:
            response = urlopen(Request(url, headers=headers))
        except HTTPError as err:
            if err.code == 401 or (err.code == 403 and 'Forbidden' in err.msg):
                msg = ('You have provided invalid authorisation credentials',)
            elif err.code == 404:
                # Page not found error.
                # For example, fetching .../releases/latest might not exist
                # if no releases for the repo are available yet. This is not
                # (necessarily) an error; however, if the API routes change
                # then this error would silently occur but must be fixed.
                return dict()
            elif err.code == 403:
                reply = fetch('/rate_limit')
                core = reply.get('resources', {}).get('core', {})
                if core.get('remaining') == 0:
                    reset = core.get('reset', -1)
                    if reset > 0:
                        hm = datetime.datetime.fromtimestamp(reset+60).strftime('%I:%M%p')
                        msg = ('GitHub rate limit exceeded. Not all of the information could be retrieved.\n'
                               'Retry after %s or create a file with your GitHub authorisation credentials to\n'
                               'increase your rate limit. Run "msl authorise --help" for more details.', hm)
                    else:
                        msg = ("%s (no 'reset' key)", err)
                else:
                    msg = ("%s (no 'remaining' key)", err)
            else:
                msg = ('Unhandled %s', err)
            fetch_errors[err.code] = msg
        except URLError:
            # Possible reasons for this error are:
            # 1. Computer has no internet access
            # 2. GitHub server is offline
            # 3. The hostname for the GitHub API has changed
            fetch_errors[0] = ('Cannot access %s', url)
        else:
            return json.loads(response.read().decode('utf-8'))

    def fetch_latest_release(repo_name):
        reply = fetch('/repos/MSLNZ/{}/releases/latest'.format(repo_name))
        if reply:
            def verify(ver):
                # ensure that the version is valid according to PEP 440
                try:
                    ver = ver.lstrip('v')
                    packaging.version.Version(ver)
                    return ver
                except packaging.version.InvalidVersion:
                    return ''
            version = verify(reply['tag_name']) or verify(reply['name'])
        else:
            version = ''
        pkgs[repo_name]['version'] = version

    def fetch_tags(repo_name):
        reply = fetch('/repos/MSLNZ/{}/tags'.format(repo_name))
        pkgs[repo_name]['tags'] = [tag['name'] for tag in reply] if reply else []

    def fetch_branches(repo_name):
        reply = fetch('/repos/MSLNZ/{}/branches'.format(repo_name))
        pkgs[repo_name]['branches'] = [branch['name'] for branch in reply] if reply else []

    # Check if the user specified their GitHub authorisation credentials.
    # The os.environ option is used for CI testing, it is not the
    # recommended way for a user to store their credentials.
    auth = os.environ.get('MSL_PM_GITHUB_AUTH')
    if not auth and os.path.isfile(_GITHUB_AUTH_PATH):
        with open(_GITHUB_AUTH_PATH, mode='rb') as fp:
            line = fp.readline().strip()
            auth = base64.b64encode(line).decode('utf-8')

    headers = {
        'User-Agent': _PKG_NAME + '/Python',
        'Accept': 'application/vnd.github+json',
    }
    if auth:
        headers['Authorization'] = 'Basic ' + auth

    log.debug('Getting the repositories from GitHub')

    fetch_errors = {}

    # when more than 100 repos exist we'll need to also fetch "?per_page=100&page=2"
    repos = fetch('/orgs/MSLNZ/repos?per_page=100&page=1')
    if not repos:
        # even though updating the cache was requested just reload the cached data
        # because GitHub cannot be connected to right now
        for error in fetch_errors.values():
            log.warning(*error)
        cache, _ = _inspect_github_pypi('github', False)
        return cache

    pkgs = dict()
    for repo in repos:
        if repo['language'] == 'Python':
            if repo['description'] is None:
                repo['description'] = ''
            pkgs[repo['name']] = {'description': repo['description']}

    threads = [
        threading.Thread(target=fcn, args=(repo_name,))
        for fcn in [fetch_latest_release, fetch_tags, fetch_branches]
        for repo_name in pkgs
    ]
    for t in threads:
        t.start()
    for t in threads:
        t.join()
    for error in fetch_errors.values():
        log.warning(*error)

    # replace the failed requests with the cached information
    if fetch_errors and os.path.isfile(path):
        with open(path, mode='rt') as fp:
            cached = json.load(fp)
        for name, value in pkgs.items():
            for item in ('version', 'tags', 'branches'):
                if not value[item] and name in cached:
                    pkgs[name][item] = cached[name][item]

    with open(path, mode='wt') as fp:
        json.dump(pkgs, fp, indent=2)

    return _sort_packages(pkgs)


def info(from_github=False, from_pypi=False, update_cache=False, as_json=False):
    """Show information about MSL packages.

    The information about the packages can be either those that are installed or
    those that are available as repositories_ on GitHub or as packages_ on PyPI.

    The default action is to show the information about the MSL packages that are installed.

    Parameters
    ----------
    from_github : :class:`bool`, optional
        Whether to show the information about the MSL repositories_ that are available on GitHub.
    from_pypi : :class:`bool`, optional
        Whether to show the information about the MSL packages_ that are available on PyPI.
    update_cache : :class:`bool`, optional
        The information about the MSL packages_ that are available on PyPI and about
        the repositories_ that are available on GitHub are cached to use for subsequent
        calls to this function. After 24 hours the cache is automatically updated. Set
        `update_cache` to be :data:`True` to force the cache to be updated when you call
        this function. If `from_github` is :data:`True` then the cache for the
        repositories_ is updated. If `from_pypi` is :data:`True` then the cache for the
        packages_ is updated.
    as_json : :class:`bool`, optional
        Whether to show the information in JSON_ format. If enabled then the information
        about the MSL repositories_ includes additional information about the branches
        and tags.
    """
    if from_github:
        typ, pkgs = 'Repository', github(update_cache=update_cache)
        if not pkgs:
            return
    elif from_pypi:
        typ, pkgs = 'PyPI Package', pypi(update_cache=update_cache)
        if not pkgs:
            return
    else:
        typ, pkgs = 'Package', installed()

    if as_json:
        if not (from_github or from_pypi):
            for name, value in pkgs.items():
                pkgs[name]['requires'] = [str(r) for r in value['requires']]
        log.debug(Fore.RESET)
        log.info(json.dumps(pkgs, indent=2))
        return

    # determine the maximum width of each column
    header = ['MSL ' + typ, 'Version', 'Description']
    w = [len(h) for h in header]
    for p in pkgs:
        w = [
            max(w[0], len(p)),
            max(w[1], len(pkgs[p]['version'])),
            max(w[2], len(pkgs[p]['description']))
        ]

    if 'PYCHARM_HOSTED' not in os.environ:
        # if not executing within the PyCharm IDE then the description
        # should wrap to the next line and begin where the description
        # on the previous line started
        term_w, term_h = _get_terminal_size()
        max_description_width = max(len(header[2]), term_w - (w[0] + w[1]) - 2)
        if w[2] > max_description_width:
            w[2] = max_description_width

    # log the results
    msg = [Fore.RESET]
    msg.append(' '.join(header[i].center(w[i]) for i in range(len(header))))
    msg.append(' '.join('-' * width for width in w))
    for p in sorted(pkgs):
        description = pkgs[p]['description']
        name_version = p.rjust(w[0]) + ' ' + pkgs[p]['version'].ljust(w[1]) + ' '
        if not description:
            msg.append(name_version)
        else:
            padding = ' ' * len(name_version)
            wrapped_lines = textwrap.wrap(description, width=w[2])
            msg.append(name_version + wrapped_lines[0])
            for line in wrapped_lines[1:]:
                msg.append(padding + line)

    log.info('\n'.join(msg))


def installed():
    """Get the information about the MSL packages that are installed.

    Returns
    -------
    :class:`dict`
        The information about the MSL packages that are installed.
    """
    log.debug('Getting the packages from %s', os.path.dirname(sys.executable))

    gh = github(update_cache=False)

    # refresh the working_set
    reload(pkg_resources)

    pkgs = {}
    for dist in pkg_resources.working_set:
        repo_name = None
        if dist.project_name in gh:  # the installed name might be different from the repo name
            repo_name = dist.project_name

        description = ''
        requires = []
        if dist.has_metadata(dist.PKG_INFO):
            for line in dist.get_metadata_lines(dist.PKG_INFO):
                if line.startswith('Summary:'):
                    description = line[8:].strip()  # can be UNKNOWN
                elif line.startswith('Home-page:') and 'github.com/MSLNZ' in line:
                    repo_name = line.split('/')[-1]
                    requires = dist.requires()
                    if description:
                        break

        if repo_name is None:
            continue

        if description == 'UNKNOWN':
            description = gh[repo_name]['description']

        pkgs[dist.project_name] = {
            'version': dist.version,
            'description': description,
            'repo_name': repo_name,
            'requires': requires,
        }

    return _sort_packages(pkgs)


def outdated_pypi_packages(msl_installed=None):
    """Check PyPI for all non-MSL packages that are outdated.

    .. versionadded:: 2.5.0

    Parameters
    ----------
    msl_installed : :class:`dict`, optional
        The MSL packages that are installed. If not specified
        then calls :func:`installed` to determine the
        installed packages.

    Returns
    -------
    :class:`dict`
        The information about the PyPI packages that are outdated.
    """
    pkgs_to_update = dict()
    log.debug('Checking PyPI for all non-MSL packages that are outdated')
    try:
        output = subprocess.check_output([sys.executable, '-m', 'pip', 'list', '--outdated'])
    except subprocess.CalledProcessError as e:
        log.error('ERROR: processing "pip list --outdated" raised -> %s', e)
        return pkgs_to_update

    lines = output.decode().splitlines()
    if not lines:
        return pkgs_to_update

    header = [item.lower() for item in lines[0].split()]
    if header[:3] != ['package', 'version', 'latest']:
        log.error('ERROR: pip has changed the structure of the outdated packages')
        return pkgs_to_update

    # ignore lines[1] since it is a bunch of '-'
    outdated_packages = [dict(zip(header, line.split())) for line in lines[2:]]

    if not msl_installed:
        msl_installed = installed()

    for msl_project_name, item in msl_installed.items():
        for outdated in outdated_packages:
            package = outdated['package']
            if package == 'pip':
                continue

            for msl_requirement in item['requires']:
                if msl_requirement.project_name.startswith(package):
                    # cannot update a package to the latest version
                    # on PyPI if the installed MSL package specifies
                    # that it only supports a specific version
                    if msl_requirement.specifier:
                        specifier = str(msl_requirement.specifier)
                        if package in pkgs_to_update:
                            if specifier not in pkgs_to_update[package]['version']:
                                # multiple version constraints
                                constraints = pkgs_to_update[package]['version']
                                if constraints[0] not in '<!=>~':
                                    # then `constraints` corresponds to an exact version
                                    # and is missing the leading "=="
                                    new_version = specifier
                                else:
                                    new_version = specifier + ',' + constraints
                                pkgs_to_update[package]['version'] = new_version
                        else:
                            pkgs_to_update[package] = {
                                'installed_version': outdated['version'],
                                'using_pypi': True,
                                'extras_require': '',
                                'version': specifier,
                                'repo_name': '',
                            }

            if package not in pkgs_to_update and package not in msl_installed:
                pkgs_to_update[package] = {
                    'installed_version': outdated['version'],
                    'using_pypi': True,
                    'extras_require': '',
                    'version': outdated['latest'],
                    'repo_name': '',
                }

    return _sort_packages(pkgs_to_update)


def pypi(update_cache=False):
    """Get the information about the MSL packages_ that are available on PyPI.

    Parameters
    ----------
    update_cache : :class:`bool`, optional
        The information about the MSL packages_ that are available on PyPI are
        cached to use for subsequent calls to this function. After 24 hours the
        cache is automatically updated. Set `update_cache` to be :data:`True`
        to force the cache to be updated when you call this function.

    Returns
    -------
    :class:`dict`
        The information about the MSL packages_ that are available on PyPI.
    """
    packages, path = _inspect_github_pypi('pypi', update_cache)
    if packages:
        return packages

    def request(endpoint):
        url = 'https://pypi.org' + endpoint
        try:
            response = urlopen(Request(url, headers=headers))
        except URLError:
            log.error('Cannot access %s', url)
        else:
            return response.read().decode('utf-8')

    def use_json_api():
        # use the JSON API as a backup way to get the package information from PyPI
        projects = ('msl-package-manager', 'msl-network', 'msl-loadlib', 'msl-io' 
                    'GTC', 'Quantity-Value')
        for project in projects:
            response = request('/pypi/{}/json'.format(project))
            if response:
                _info = json.loads(response).get('info')
                if _info:
                    pkgs[project] = {
                        'version': _info.get('version', 'UNKNOWN'),
                        'description': _info.get('summary', 'UNKNOWN')
                    }

    pkgs = dict()
    log.debug('Getting the packages from PyPI')
    headers = {'User-Agent': _PKG_NAME + '/Python'}

    # use the /search endpoint before the /json endpoint since searching does not
    # depend on knowing in advance what MSL packages are available on PyPI
    reply = request('/search/?q=%22Measurement+Standards+Laboratory+of+New+Zealand%22&o=')
    if reply:
        items = re.findall(
            r'<span class="package-snippet__name">(?P<name>.+)</span>\s+'
            r'<span class="package-snippet__version">(?P<version>.+)</span>\s+'
            r'<span class="package-snippet__created">.+\s+(?P<created>.+)\s+.+\s+.+\s+'
            r'<p class="package-snippet__description">(?P<description>.+)</p>',
            reply,
        )
        for item in items:
            name, version, created, description = item
            pkgs[name] = {
                'version': version,
                'description': description,
            }
        if not pkgs:
            log.warning('The regex pattern is invalid for the /search route, using /json instead')
            use_json_api()
    else:
        use_json_api()

    if not pkgs:
        return _inspect_github_pypi('pypi', False)[0]

    with open(path, mode='wt') as fp:
        json.dump(pkgs, fp, indent=2)

    return _sort_packages(pkgs)


def set_log_level(level):
    """Set the logging :py:ref:`level <levels>`.

    Parameters
    ----------
    level : :class:`int`
        A value from one of the :py:ref:`levels`.
    """
    global _pip_quiet

    if level <= logging.INFO:
        _pip_quiet = 0
    elif level == logging.WARNING:
        _pip_quiet = 1
    elif level == logging.ERROR:
        _pip_quiet = 2
    else:
        _pip_quiet = 3

    log.setLevel(level)


def _ask_proceed():
    """Ask whether to proceed with the command.

    Returns
    -------
    :class:`bool`
        Whether to proceed.
    """
    ask = '\nProceed (Y/n)? '
    response = _get_input(ask).lower()
    while True:
        if response.startswith('y') or not response:
            return True
        elif response.startswith('n'):
            return False
        else:
            log.info('Invalid response')
            response = _get_input(ask).lower()


def _check_kwargs(kwargs, allowed):
    for item in kwargs:
        if item not in allowed:
            log.warning('Invalid kwarg %r', item)


def _check_wildcards_and_prefix(names, pkgs):
    """Check if a Unix shell-style wildcard was used or if the package
    name starts with the msl- prefix.

    Parameters
    ----------
    names : :class:`tuple` of :class:`str`
       The package names.
    pkgs : :class:`dict`
       The GitHub repositories or the installed packages.

    Returns
    -------
    :class:`dict`
       The keys are the package names and the values are the version info/extra requires.
    """
    _packages = dict()
    pkgs_map = dict((p.lower(), p) for p in pkgs)
    for name in names:
        found = re.search(_package_name_regex, name.lower())
        if not found:
            continue
        result = found.groupdict()
        if not result['package_name']:
            log.error('Invalid package name %r', name)
            continue
        found_it = False
        for prefix in ['', 'msl-']:
            for match in fnmatch.filter(pkgs_map, prefix+result['package_name']):
                found_it = True
                _packages[pkgs_map[match]] = {
                    'extras_require': result['extras_require'],
                    'version_requested': result['version_requested']
                }
        if not found_it:
            log.warning('No MSL packages match %r', name)
    return _packages


def _create_install_list(names, branch, commit, tag, update_cache):
    """Create a list of package names to ``install`` that are GitHub repositories_.

    Parameters
    ----------
    names : :class:`tuple` of :class:`str`
        The name of a single GitHub repository or multiple repository names.
    branch : :class:`str` or :data:`None`
        The name of a git branch.
    commit : :class:`str` or :data:`None`
        The hash value of a git commit.
    tag : :class:`str` or :data:`None`
        The name of a git tag.
    update_cache : :class:`bool`
        Whether to force the GitHub cache to be updated when you call this function.

    Returns
    -------
    :class:`dict`
        The MSL packages to ``install``.
    """
    github_suffix = _get_github_url_suffix(branch, commit, tag)
    if github_suffix is None:
        return

    # keep the order of the log messages consistent: pypi -> github -> local
    pkgs_github = github(update_cache=update_cache)
    pkgs_installed = installed()

    if not names:  # e.g., the --all flag
        packages = dict((pkg, {'extras_require': None, 'version_requested': None})
                        for pkg in pkgs_github if pkg != _PKG_NAME and pkg.startswith('msl-'))
    else:
        packages = _check_wildcards_and_prefix(names, pkgs_github)

    # the name of an installed package can be different from the repo name
    repo_names = [p['repo_name'] for p in pkgs_installed.values()]

    pkgs = {}
    for name, value in packages.items():
        if name in pkgs_installed or name in repo_names:
            log.warning('The %r package is already installed -- use the update command', name)
        elif name not in pkgs_github:
            log.error('Cannot install %r -- the package does not exist', name)
        elif branch is not None and branch not in pkgs_github[name]['branches']:
            log.error('Cannot install %r -- a %r branch does not exist', name, branch)
        elif tag is not None and tag not in pkgs_github[name]['tags']:
            log.error('Cannot install %r -- a %r tag does not exist', name, tag)
        else:
            pkgs[name] = pkgs_github[name]
            pkgs[name]['version_requested'] = value['version_requested']
            pkgs[name]['extras_require'] = value['extras_require']
    return pkgs


def _create_uninstall_list(names):
    """Create a list of package names to ``uninstall``.

    Parameters
    ----------
    names : :class:`tuple` of :class:`str`
        The name(s) of the package(s) to ``uninstall``.

    Returns
    -------
    :class:`dict`
        The MSL packages to ``uninstall``.
    """

    pkgs_installed = installed()

    if not names:
        names = [pkg for pkg in pkgs_installed if pkg != _PKG_NAME]
    else:
        names = _check_wildcards_and_prefix(names, pkgs_installed)

    pkgs = {}
    for name in names:
        if name == _PKG_NAME:
            log.warning('The MSL Package Manager cannot uninstall itself, '
                        'run "pip uninstall %s"', _PKG_NAME)
        elif name not in pkgs_installed:
            log.error('Cannot uninstall %r -- the package is not installed.', name)
        else:
            pkgs[name] = pkgs_installed[name]
    return pkgs


def _get_input(msg):
    """Get input from the user.

    Parameters
    ----------
    msg : :class:`str`
        The message to display.

    Returns
    -------
    :class:`str`
        The user's input from :attr:`sys.stdin`.
    """
    try:
        return raw_input(msg)  # Python 2
    except NameError:
        return input(msg)


def _get_github_url_suffix(branch=None, commit=None, tag=None):
    """Returns the URL suffix.

    Parameters
    ----------
    branch : :class:`str` or :data:`None`
        The name of a git branch.
    commit : :class:`str` or :data:`None`
        The hash value of a git commit.
    tag : :class:`str` or :data:`None`
        The name of a git tag.

    Returns
    -------
    :class:`str` or :data:`None`
        The suffix or :data:`None` if more than 1 of
        `branch`, `tag` or `commit` were enabled.
    """
    count = [bool(branch), bool(commit), bool(tag)].count(True)
    if count == 0:
        return 'main'  # default branch

    if count > 1:
        log.error('Can only specify a branch, a commit or a tag (not multiple options simultaneously)')
        return

    if branch:
        return branch

    if commit:
        return commit

    return tag


def _inspect_github_pypi(where, update_cache):
    """Inspects the HOME directory for the cached json file.

    Parameters
    ----------
    where : :class:`str`
        Either 'github' or 'pypi'.
    update_cache : :class:`bool`
        Whether to update the cache.

    Returns
    -------
    :class:`dict`
        The packages.
    :class:`str`
        The path to the cached file.
    """
    if where == 'github':
        suffix = 'GitHub repositories'
    elif where == 'pypi':
        suffix = 'PyPI packages'
    else:
        assert False, '{!r} != github or pypi'.format(where)

    path = os.path.join(_HOME_DIR, where+'.json')

    cached_pgks = None
    if os.path.isfile(path):
        with open(path, mode='rt') as f:
            cached_pgks = _sort_packages(json.load(f))

    one_day = 60 * 60 * 24
    if (not update_cache) and (cached_pgks is not None) and (time.time() < os.path.getmtime(path) + one_day):
        # The installed() function also calls github() so this log message could be displayed twice.
        # Avoid seeing the following log message when the installed() function was previously called.
        if where == 'pypi' or not _ColourStreamHandler.previous_message.endswith(os.path.dirname(sys.executable)):
            log.debug('Loaded the cached information about the %s', suffix)
        return _sort_packages(cached_pgks), path

    return dict(), path


def _log_install_uninstall_message(packages, action, branch=None, commit=None, tag=None, pkgs_pypi=None):
    """Print the ``install`` or ``uninstall`` summary for what is going to happen.

    Parameters
    ----------
    packages : :class:`dict`
        The packages that are affected.
    action : :class:`str`
        The text to show in color and in upper case about what's happening.
    branch : :class:`str` or :data:`None`
        The name of a git branch to use. Only used when installing packages.
    commit : :class:`str` or :data:`None`
        The hash value of a git commit to use. Only used when installing packages.
    tag : :class:`str` or :data:`None`
        The name of a git tag to use. Only used when installing packages.
    pkgs_pypi : :class:`dict`
        Packages that are on PyPI. Only used when installing packages.
    """
    pkgs = _sort_packages(packages)

    w = [0, 0]
    for pkg, values in pkgs.items():
        if values.get('extras_require'):
            w[0] = max(w[0], len(pkg + values['extras_require']))
        else:
            w[0] = max(w[0], len(pkg))
        if values.get('version_requested'):
            w[1] = max(w[1], len(values['version_requested']))
        else:
            w[1] = max(w[1], len(values['version']))

    msg = '\n{}The following MSL packages will be {}{}{}:\n'.format(Fore.RESET, Fore.CYAN, action, Fore.RESET)
    for pkg, values in pkgs.items():
        name = pkg
        if values.get('extras_require'):
            name += values['extras_require']
        if values.get('version_requested'):
            version = values['version_requested'].replace('==', '')
        else:
            version = '' if (branch or commit or tag) else values['version']
        msg += '\n  {}  {} '.format(name.ljust(w[0]), version.ljust(w[1]))
        if action == 'REMOVED':
            continue
        if branch is not None:
            msg += ' [branch:{}]'.format(branch)
        elif commit is not None:
            msg += ' [commit:{}]'.format(commit[:7])
        elif tag is not None:
            msg += ' [tag:{}]'.format(tag)
        elif pkg in pkgs_pypi:
            msg += ' [PyPI]'
        else:
            msg += ' [GitHub]'

    log.info(msg)


def _sort_packages(pkgs):
    """Sort the MSL packages by the name of the package.

    Parameters
    ----------
    pkgs : :class:`dict`
        The MSL packages.

    Returns
    -------
    :class:`collections.OrderedDict`
        The packages sorted by name.
    """
    return collections.OrderedDict([(u'{}'.format(k), pkgs[k]) for k in sorted(pkgs)])


class _ColourStreamHandler(logging.StreamHandler):
    """A SteamHandler that is compatible with colorama."""

    COLOURS = {
        'DEBUG': Fore.CYAN,
        'INFO': '',
        'WARN': Fore.YELLOW,
        'WARNING': Fore.YELLOW,
        'ERROR': Style.BRIGHT + Fore.RED,
        'CRIT': Back.RED + Fore.WHITE,
        'CRITICAL': Back.RED + Fore.WHITE
    }

    previous_message = ''

    def emit(self, record):
        _ColourStreamHandler.previous_message = record.getMessage()
        stream = sys.stdout if record.levelno < logging.WARNING else sys.stderr
        try:
            stream.write(self.COLOURS[record.levelname] + self.format(record) + '\n')
            stream.flush()
        except:
            self.handleError(record)


def _getLogger(name=None, fmt='%(message)s'):
    """Create the default stream logger"""
    init(autoreset=True)  # initialize colorama
    logger = logging.getLogger(name)
    handler = _ColourStreamHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(fmt))
    logger.addHandler(handler)
    logger.setLevel(logging.DEBUG)
    return logger


log = _getLogger(_PKG_NAME)


def _get_terminal_size():
    """
    Taken from https://gist.github.com/jtriley/1108174

    getTerminalSize()
     - get width and height of console
     - works on linux,os x,windows,cygwin(windows)

     originally retrieved from:
     https://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
    """
    current_os = platform.system()
    tuple_xy = None
    if current_os == 'Windows':
        tuple_xy = _get_terminal_size_windows()
        if tuple_xy is None:
            tuple_xy = _get_terminal_size_tput()
            # needed for window's python in cygwin's xterm!
    if current_os in ['Linux', 'Darwin'] or current_os.startswith('CYGWIN'):
        tuple_xy = _get_terminal_size_linux()
    if tuple_xy is None or tuple_xy == (0, 0):
        tuple_xy = (80, 25)  # default value
    return tuple_xy


def _get_terminal_size_windows():
    try:
        from ctypes import windll, create_string_buffer
        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12
        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
        if res:
            (bufx, bufy, curx, cury, wattr,
             left, top, right, bottom,
             maxx, maxy) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            return right - left, bottom - top
    except:
        pass


def _get_terminal_size_tput():
    # https://stackoverflow.com/questions/263890/how-do-i-find-the-width-height-of-a-terminal-window
    try:
        cols = int(subprocess.check_call(shlex.split('tput cols')))
        rows = int(subprocess.check_call(shlex.split('tput lines')))
        return cols, rows
    except:
        pass


def _get_terminal_size_linux():
    def ioctl_GWINSZ(fd):
        try:
            import fcntl
            import termios
            cr = struct.unpack('hh', fcntl.ioctl(fd, termios.TIOCGWINSZ, '1234'))
            return cr
        except:
            pass

    cr = ioctl_GWINSZ(0) or ioctl_GWINSZ(1) or ioctl_GWINSZ(2)
    if not cr:
        try:
            fd = os.open(os.ctermid(), os.O_RDONLY)
            cr = ioctl_GWINSZ(fd)
            os.close(fd)
        except:
            pass
    if not cr:
        try:
            cr = (os.environ['LINES'], os.environ['COLUMNS'])
        except:
            return None
    return int(cr[1]), int(cr[0])



import os
import shutil
import tempfile
from datetime import datetime

import pytest

import msl.package_manager as pm

temp_dir = tempfile.gettempdir()


def test_invalid_kwarg_types():
    with pytest.raises(TypeError):
        pm.create('abc', author=1)

    with pytest.raises(TypeError):
        pm.create('abc', author='a', email=['a'])


def test_msl_package():
    root_dir = os.path.join(temp_dir, 'msl-package')
    if os.path.isdir(root_dir):
        shutil.rmtree(root_dir)

    pm.create('package', author='Abc Def', email='a.b@c.com', directory=temp_dir)

    assert os.path.isdir(root_dir)
    assert os.path.isdir(os.path.join(root_dir, 'msl'))
    assert os.path.isdir(os.path.join(root_dir, 'msl', 'package'))
    assert os.path.isdir(os.path.join(root_dir, 'msl', 'examples'))
    assert os.path.isdir(os.path.join(root_dir, 'msl', 'examples', 'package'))
    assert os.path.isdir(os.path.join(root_dir, 'docs'))
    assert os.path.isdir(os.path.join(root_dir, 'docs', '_static'))
    assert os.path.isdir(os.path.join(root_dir, 'tests'))
    assert os.path.isfile(os.path.join(root_dir, '.coveragerc'))
    assert os.path.isfile(os.path.join(root_dir, '.gitignore'))
    assert os.path.isfile(os.path.join(root_dir, 'AUTHORS.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'CHANGES.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'condatests.py'))
    assert os.path.isfile(os.path.join(root_dir, 'LICENSE.txt'))
    assert os.path.isfile(os.path.join(root_dir, 'MANIFEST.in'))
    assert os.path.isfile(os.path.join(root_dir, 'README.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'setup.cfg'))
    assert os.path.isfile(os.path.join(root_dir, 'setup.py'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'authors.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'changelog.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'conf.py'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'index.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'license.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'tests', 'README'))

    with open(os.path.join(root_dir, 'msl', '__init__.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert 'DO NOT EDIT THIS FILE.' == lines[1]
        assert lines[3].endswith('**msl** namespace.')
        assert "__path__ = __import__('pkgutil').extend_path(__path__, __name__)" == lines[5]

    with open(os.path.join(root_dir, 'msl', 'package', '__init__.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert "__author__ = 'Measurement Standards Laboratory of New Zealand'" == lines[3]
        assert "__copyright__ = '\\xa9 {}, ' + __author__".format(datetime.now().strftime('%Y')) == lines[4]

    with open(os.path.join(root_dir, 'msl', 'examples', '__init__.py'), mode='rt') as fp:
        assert '# DO NOT EDIT THIS FILE.\n' == fp.read()

    with open(os.path.join(root_dir, 'msl', 'examples', 'package', '__init__.py'), mode='rt') as fp:
        assert '' == fp.read()

    with open(os.path.join(root_dir, '.coveragerc'), mode='rt') as fp:
        lines = fp.read()
        assert 'omit =\n    msl/examples/*' in lines
        assert lines.endswith('title = msl-package coverage report\n')

    with open(os.path.join(root_dir, 'AUTHORS.rst'), mode='rt') as fp:
        assert '* Abc Def <a.b@c.com>' in fp.read()

    with open(os.path.join(root_dir, 'README.rst'), mode='rt') as fp:
        assert fp.read().startswith('msl-package\n===========\n')

    with open(os.path.join(root_dir, 'setup.cfg'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert '    --cov msl' == lines[11]
        assert '    --ignore msl/examples' == lines[19]

    with open(os.path.join(root_dir, 'setup.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert "'msl',  # the path to" in lines[31]
        assert "# specify the packages that msl-package depends on" == lines[146]
        assert "init_original = 'msl/package/__init__.py'" == lines[161]
        assert "    name='msl-package'," == lines[166]
        assert "    url='https://github.com/MSLNZ/msl-package'," == lines[170]
        assert "    description='Write a short description about msl-package here'," == lines[171]
        assert "    packages=find_packages(include=('msl*',))," == lines[185]

    with open(os.path.join(root_dir, 'docs', 'conf.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert 'from msl import package' == lines[10]
        assert "project = 'msl-package'" == lines[75]
        assert lines[76].startswith("copyright = package.__copyright__")
        assert 'author = package.__author__' == lines[77]
        assert 'version = package.__version__' == lines[84]
        assert 'release = package.__version__' == lines[86]
        assert 'msl-package.tex' in lines[160]

    with open(os.path.join(root_dir, 'docs', 'index.rst'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert '.. _msl-package-welcome:' == lines[0]
        assert 'msl-package' == lines[2]
        assert '===========' == lines[3]

    shutil.rmtree(root_dir)


def test_pr_single_photons():
    root_dir = os.path.join(temp_dir, 'pr-Single-Photons')
    if os.path.isdir(root_dir):
        shutil.rmtree(root_dir)

    for namespace in ['pr', 'pr-', 'PR----']:
        pm.create('Single-Photons', author='joe', email='joe@email.com', namespace=namespace, directory=temp_dir)

        assert os.path.isdir(root_dir)
        assert os.path.isdir(os.path.join(root_dir, 'pr'))
        assert os.path.isdir(os.path.join(root_dir, 'pr', 'Single_Photons'))
        assert os.path.isdir(os.path.join(root_dir, 'pr', 'examples'))
        assert os.path.isdir(os.path.join(root_dir, 'pr', 'examples', 'Single_Photons'))
        assert os.path.isdir(os.path.join(root_dir, 'docs'))
        assert os.path.isdir(os.path.join(root_dir, 'docs', '_static'))
        assert os.path.isdir(os.path.join(root_dir, 'tests'))
        assert os.path.isfile(os.path.join(root_dir, '.coveragerc'))
        assert os.path.isfile(os.path.join(root_dir, '.gitignore'))
        assert os.path.isfile(os.path.join(root_dir, 'AUTHORS.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'CHANGES.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'condatests.py'))
        assert os.path.isfile(os.path.join(root_dir, 'LICENSE.txt'))
        assert os.path.isfile(os.path.join(root_dir, 'MANIFEST.in'))
        assert os.path.isfile(os.path.join(root_dir, 'README.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'setup.cfg'))
        assert os.path.isfile(os.path.join(root_dir, 'setup.py'))
        assert os.path.isfile(os.path.join(root_dir, 'docs', 'authors.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'docs', 'changelog.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'docs', 'conf.py'))
        assert os.path.isfile(os.path.join(root_dir, 'docs', 'index.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'docs', 'license.rst'))
        assert os.path.isfile(os.path.join(root_dir, 'tests', 'README'))

        with open(os.path.join(root_dir, 'pr', '__init__.py'), mode='rt') as fp:
            lines = [line.rstrip() for line in fp.readlines()]
            assert 'DO NOT EDIT THIS FILE.' == lines[1]
            assert lines[3].endswith('**pr** namespace.')
            assert "__path__ = __import__('pkgutil').extend_path(__path__, __name__)" == lines[5]

        with open(os.path.join(root_dir, 'pr', 'Single_Photons', '__init__.py'), mode='rt') as fp:
            lines = [line.rstrip() for line in fp.readlines()]
            assert "__author__ = 'Measurement Standards Laboratory of New Zealand'" == lines[3]
            assert "__copyright__ = '\\xa9 {}, ' + __author__".format(datetime.now().strftime('%Y')) == lines[4]

        with open(os.path.join(root_dir, 'pr', 'examples', '__init__.py'), mode='rt') as fp:
            assert '# DO NOT EDIT THIS FILE.\n' == fp.read()

        with open(os.path.join(root_dir, 'pr', 'examples', 'Single_Photons', '__init__.py'), mode='rt') as fp:
            assert '' == fp.read()

        with open(os.path.join(root_dir, '.coveragerc'), mode='rt') as fp:
            lines = fp.read()
            assert 'omit =\n    pr/examples/*' in lines
            assert lines.endswith('title = pr-Single-Photons coverage report\n')

        with open(os.path.join(root_dir, 'AUTHORS.rst'),  mode='rt') as fp:
            assert '* joe <joe@email.com>' in fp.read()

        with open(os.path.join(root_dir, 'README.rst'), mode='rt') as fp:
            assert fp.read().startswith('pr-Single-Photons\n=================\n')

        with open(os.path.join(root_dir, 'setup.cfg'), mode='rt') as fp:
            lines = [line.rstrip() for line in fp.readlines()]
            assert '    --cov pr' == lines[11]
            assert '    --ignore pr/examples' == lines[19]

        with open(os.path.join(root_dir, 'setup.py'), mode='rt') as fp:
            lines = [line.rstrip() for line in fp.readlines()]
            assert "'pr',  # the path to" in lines[31]
            assert "# specify the packages that pr-Single-Photons depends on" == lines[146]
            assert "init_original = 'pr/Single_Photons/__init__.py'" == lines[161]
            assert "    name='pr-Single-Photons'," == lines[166]
            assert "    url='https://github.com/MSLNZ/pr-Single-Photons'," == lines[170]
            assert "    description='Write a short description about pr-Single-Photons here'," == lines[171]
            assert "    packages=find_packages(include=('pr*',))," == lines[185]

        with open(os.path.join(root_dir, 'docs', 'conf.py'), mode='rt') as fp:
            lines = [line.rstrip() for line in fp.readlines()]
            assert 'from pr import Single_Photons' == lines[10]
            assert "project = 'pr-Single-Photons'" == lines[75]
            assert lines[76].startswith("copyright = Single_Photons.__copyright__")
            assert 'author = Single_Photons.__author__' == lines[77]
            assert 'version = Single_Photons.__version__' == lines[84]
            assert 'release = Single_Photons.__version__' == lines[86]
            assert 'pr-Single-Photons.tex' in lines[160]

        with open(os.path.join(root_dir, 'docs', 'index.rst'), mode='rt') as fp:
            lines = [line.rstrip() for line in fp.readlines()]
            assert '.. _pr-Single-Photons-welcome:' == lines[0]
            assert 'pr-Single-Photons' == lines[2]
            assert '=================' == lines[3]

        shutil.rmtree(root_dir)


def test_no_namespace():
    root_dir = os.path.join(temp_dir, 'time-tagger')
    if os.path.isdir(root_dir):
        shutil.rmtree(root_dir)

    pm.create('time-tagger', author=['Time', 'T.', 'Tagger'],
              email='time@tagger.com', directory=temp_dir, namespace=None)

    assert os.path.isdir(root_dir)
    assert os.path.isdir(os.path.join(root_dir, 'time_tagger'))
    assert os.path.isdir(os.path.join(root_dir, 'docs'))
    assert os.path.isdir(os.path.join(root_dir, 'docs', '_static'))
    assert os.path.isdir(os.path.join(root_dir, 'tests'))
    assert os.path.isfile(os.path.join(root_dir, '.coveragerc'))
    assert os.path.isfile(os.path.join(root_dir, '.gitignore'))
    assert os.path.isfile(os.path.join(root_dir, 'AUTHORS.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'CHANGES.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'condatests.py'))
    assert os.path.isfile(os.path.join(root_dir, 'LICENSE.txt'))
    assert os.path.isfile(os.path.join(root_dir, 'MANIFEST.in'))
    assert os.path.isfile(os.path.join(root_dir, 'README.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'setup.cfg'))
    assert os.path.isfile(os.path.join(root_dir, 'setup.py'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'authors.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'changelog.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'conf.py'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'index.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'docs', 'license.rst'))
    assert os.path.isfile(os.path.join(root_dir, 'tests', 'README'))

    with open(os.path.join(root_dir, 'time_tagger', '__init__.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert "__author__ = 'Measurement Standards Laboratory of New Zealand'" == lines[3]
        assert "__copyright__ = '\\xa9 {}, ' + __author__".format(datetime.now().strftime('%Y')) == lines[4]

    with open(os.path.join(root_dir, '.coveragerc'), mode='rt') as fp:
        lines = fp.read()
        assert 'omit =' not in lines
        assert lines.endswith('title = time-tagger coverage report\n')

    with open(os.path.join(root_dir, 'AUTHORS.rst'), mode='rt') as fp:
        assert '* Time T. Tagger <time@tagger.com>' in fp.read()

    with open(os.path.join(root_dir, 'README.rst'), mode='rt') as fp:
        assert fp.read().startswith('time-tagger\n===========\n')

    with open(os.path.join(root_dir, 'setup.cfg'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert '    --cov time_tagger' == lines[11]
        assert '    --ignore time_tagger/examples' != lines[19]

    with open(os.path.join(root_dir, 'setup.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert "'time_tagger',  # the path to" in lines[31]
        assert "# specify the packages that time-tagger depends on" == lines[146]
        assert "init_original = 'time_tagger/__init__.py'" == lines[161]
        assert "    name='time-tagger'," == lines[166]
        assert "    url='https://github.com/MSLNZ/time-tagger'," == lines[170]
        assert "    description='Write a short description about time-tagger here'," == lines[171]
        assert "    packages=find_packages(include=('time_tagger',))," == lines[185]

    with open(os.path.join(root_dir, 'docs', 'conf.py'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert 'import time_tagger' == lines[10]
        assert "project = 'time-tagger'" == lines[75]
        assert lines[76].startswith("copyright = time_tagger.__copyright__")
        assert 'author = time_tagger.__author__' == lines[77]
        assert 'version = time_tagger.__version__' == lines[84]
        assert 'release = time_tagger.__version__' == lines[86]
        assert 'time-tagger.tex' in lines[160]

    with open(os.path.join(root_dir, 'docs', 'index.rst'), mode='rt') as fp:
        lines = [line.rstrip() for line in fp.readlines()]
        assert '.. _time-tagger-welcome:' == lines[0]
        assert 'time-tagger' == lines[2]
        assert '===========' == lines[3]

    shutil.rmtree(root_dir)



"""
Create an authorisation file for the GitHub API.
"""
from .utils import _GITHUB_AUTH_PATH
from .utils import _get_input
from .utils import get_username
from .utils import log

WARNING_MESSAGE = """
Your username and personal access token are saved in plain text in the
file that is created. You should set the file permissions provided by 
your operating system to ensure that your GitHub credentials are safe.
"""


def authorise(username=None, token=None):
    """Create an authorisation file for the GitHub API.

    When requesting information about the MSL repositories_ that are
    available on GitHub there is a limit to how often you can send
    requests to the GitHub API. If you have a GitHub account and
    include your username and a `personal access token`_ with each
    request then this limit is increased.

    .. important::

       Calling this function will create a file that contains your GitHub
       username and a `personal access token`_ so that GitHub requests are
       authorised. Your username and `personal access token`_ are saved in
       plain text in the file that is created. You should set the file
       permissions provided by your operating system to ensure that your
       GitHub credentials are safe.

    .. versionadded:: 2.3.0

    .. versionchanged:: 2.4.0
        Renamed the `password` keyword argument to `token`.

    .. versionchanged:: 2.5.0
        Renamed function to `authorise`.

    .. _repositories: https://github.com/MSLNZ
    .. _personal access token: https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token

    Parameters
    ----------
    username : :class:`str`, optional
        The GitHub username. If :data:`None` then you will be
        asked for the `username`.
    token : :class:`str`, optional
        A GitHub `personal access token`_ for `username`. If :data:`None`
        then you will be asked for the `token`.
    """
    if username is None:
        default = get_username()
        try:
            username = _get_input('Enter your GitHub username [default: {}]: '.format(default))
        except KeyboardInterrupt:
            log.warning('\nDid not create the GitHub authorisation file')
            return
        else:
            if not username:
                username = default

    if token is None:
        try:
            token = _get_input('Enter your GitHub personal access token: ')
        except KeyboardInterrupt:
            log.warning('\nDid not create the GitHub authorisation file')
            return

    if not username:
        log.warning('You must enter a username. Did not create the GitHub authorisation file')
        return

    if not token:
        log.warning('You must enter a personal access token. Did not create the GitHub authorisation file')
        return

    with open(_GITHUB_AUTH_PATH, mode='wt') as fp:
        fp.write(username + ':' + token)

    log.warning(WARNING_MESSAGE)
    log.info('GitHub credentials were saved to %s', _GITHUB_AUTH_PATH)



import logging
import os
import sys

import pytest

from msl.package_manager import install
from msl.package_manager import installed
from msl.package_manager import uninstall
from msl.package_manager import update
from msl.package_manager import utils


# see the docs for the caplog fixture
# https://docs.pytest.org/en/latest/logging.html#caplog-fixture
def test_log_output(caplog):
    # checks that the logging messages have the expected output
    caplog.set_level(logging.DEBUG, logger=utils._PKG_NAME)

    # install MSL-LoadLib version 0.5.0
    install('loadlib==0.5.0', yes=True)

    # make sure that MSL-LoadLib is installed
    assert 'msl-loadlib' in installed()

    # install a package that does not exist
    install('does_not_exist', yes=True)

    # install MSL-LoadLib
    install('loadlib', yes=True)

    # check that py4j is not installed
    with pytest.raises(ImportError):
        import py4j

    # update MSL-LoadLib
    update('loadlib[java]==0.6.0', yes=True)

    # py4j should now be installed
    import py4j

    # update a package that is not an MSL package
    update('colorama', yes=True)

    # uninstall a package that is not an MSL package
    uninstall('colorama', yes=True)
    import colorama  # still installed

    # install a package that is not part of the msl namespace
    install('GTC<1.3', yes=True, pip_options=['--no-deps'])

    # make sure that GTC is installed
    assert 'GTC' in installed()

    # update GTC -> invalid branch
    update('GTC', yes=True, branch='invalid')

    # update GTC -> invalid tag
    update('GTC', yes=True, tag='invalid')

    # update GTC -> tag=v1.3.1
    update('GTC', yes=True, tag='v1.3.1')

    # uninstall GTC
    uninstall('GTC', yes=True)

    # make sure that GTC is not installed but msl-loadlib is still installed
    assert 'GTC' not in installed()
    assert 'msl-loadlib' in installed()

    # uninstall MSL-LoadLib
    uninstall('loadlib', yes=True)

    # make sure that MSL-LoadLib is not installed
    assert 'msl-loadlib' not in installed()

    #
    # the expected logging messages
    #
    exec_path = os.path.dirname(sys.executable)
    u = 'u' if sys.version_info.major == 2 else ''
    expected = [
        # msl install loadlib==0.5.0
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        '\n\x1b[39mThe following MSL packages will be \x1b[36mINSTALLED\x1b[39m:\n\n  msl-loadlib  0.5.0    [PyPI]',
        '',
        "Installing {}'msl-loadlib' from PyPI".format(u),

        # check if msl-loadlib is installed
        'Getting the packages from {}'.format(exec_path),

        # msl install does_not_exist -> a that package does not exist
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        "No MSL packages match 'does_not_exist'",
        'No MSL packages to install',

        # msl install loadlib -> already installed
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        "The {}'msl-loadlib' package is already installed -- use the update command".format(u),
        'No MSL packages to install',

        # msl update loadlib[java]==0.6.0
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        '\n\x1b[39mThe following MSL packages will be \x1b[36mUPDATED\x1b[39m:\n\n  msl-loadlib[java]  0.5.0 --> 0.6.0    [PyPI]',
        '',
        "Updating {}'msl-loadlib' from PyPI".format(u),

        # msl update colorama -> not an MSL package
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        "No MSL packages match 'colorama'",
        '\x1b[39mNo packages to update\x1b[39m',

        # msl uninstall colorama -> not an MSL package
        'Getting the packages from {}'.format(exec_path),
        "No MSL packages match 'colorama'",
        'No MSL packages to uninstall',

        # install GTC
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        '\n\x1b[39mThe following MSL packages will be \x1b[36mINSTALLED\x1b[39m:\n\n  GTC  <1.3  [PyPI]',
        '',
        "Installing {}'GTC' from PyPI".format(u),

        # check if GTC is installed
        'Getting the packages from {}'.format(exec_path),

        # update GTC -> invalid branch
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        "Cannot update {}'GTC' -- the 'invalid' branch does not exist".format(u),
        '\x1b[39mNo packages to update\x1b[39m',

        # update GTC -> invalid tag
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        "Cannot update {}'GTC' -- the 'invalid' tag does not exist".format(u),
        '\x1b[39mNo packages to update\x1b[39m',

        # update GTC -> tag=v1.3.1
        'Loaded the cached information about the PyPI packages',
        'Loaded the cached information about the GitHub repositories',
        'Getting the packages from {}'.format(exec_path),
        '\n\x1b[39mThe following MSL packages will be \x1b[36mUPDATED\x1b[39m:\n\n  GTC  1.2.1 --> [tag:v1.3.1]  [GitHub]',
        '',
        "Updating {}'GTC' from GitHub[v1.3.1]".format(u),

        # msl uninstall GTC
        'Getting the packages from {}'.format(exec_path),
        '\n\x1b[39mThe following MSL packages will be \x1b[36mREMOVED\x1b[39m:\n\n  GTC  1.3.1 ',
        '',

        # checking that GTC is not installed
        'Getting the packages from {}'.format(exec_path),

        # checking that msl-loadlib is still installed
        'Getting the packages from {}'.format(exec_path),

        # msl uninstall loadlib
        'Getting the packages from {}'.format(exec_path),
        '\n\x1b[39mThe following MSL packages will be \x1b[36mREMOVED\x1b[39m:\n\n  msl-loadlib  0.6.0 ',
        '',

        # check that msl-loadlib is not installed
        'Getting the packages from {}'.format(exec_path),
    ]

    for index, record in enumerate(caplog.records):
        assert expected[index] == record.message


def test_branch_commit_tag(caplog):
    caplog.set_level(logging.DEBUG, logger=utils._PKG_NAME)

    options = [
        (False, True, True),
        (True, False, True),
        (True, True, False),
        (True, True, True),
    ]

    caplog.clear()
    for b, c, t in options:
        install('does-not-matter', branch=b, commit=c, tag=t)
        update('does-not-matter', branch=b, commit=c, tag=t)
    for r in caplog.records:
        assert r.levelname == 'ERROR'
        assert r.message == 'Can only specify a branch, a commit or a tag (not multiple options simultaneously)'

    cached_has_git = bool(utils.has_git)
    utils.has_git = False

    caplog.clear()
    install('does-not-matter', commit='does-not-matter')
    r = caplog.records[0]
    assert r.levelname == 'ERROR'
    assert r.message == 'Cannot install from a commit because git is not installed'

    caplog.clear()
    update('does-not-matter', commit='does-not-matter')
    r = caplog.records[0]
    assert r.levelname == 'ERROR'
    assert r.message == 'Cannot update from a commit because git is not installed'

    utils.has_git = cached_has_git



import logging
import os
import sys

from msl.package_manager import utils


# see the docs for the caplog fixture
# https://docs.pytest.org/en/latest/logging.html#caplog-fixture
def test_info_log(caplog):
    # checks that the logging messages have the expected output
    caplog.set_level(logging.DEBUG, logger=utils._PKG_NAME)
    exec_path = os.path.dirname(sys.executable)

    utils.info()
    expected = [
        'Getting the packages from {}'.format(exec_path),
    ]
    for index, message in enumerate(expected):
        assert caplog.records[index].message == message
    del caplog.records[:]

    utils.info(from_github=True)
    expected = [
        'Loaded the cached information about the GitHub repositories',
    ]
    for index, message in enumerate(expected):
        assert caplog.records[index].message == message
    del caplog.records[:]

    utils.info(from_pypi=True)
    expected = [
        'Loaded the cached information about the PyPI packages',
    ]
    for index, message in enumerate(expected):
        assert caplog.records[index].message == message
    del caplog.records[:]


def test_create_install_list():

    # args -> names, branch, tag, update_cache

    pkgs = utils._create_install_list((), None, None, None, False)
    for p in pkgs:
        assert p.startswith('msl-')
    assert utils._PKG_NAME not in pkgs

    # "msl-" prefix gets added
    pkgs = utils._create_install_list(('loadlib', 'network'), None, None, None, False)
    assert len(pkgs) == 2
    assert 'msl-loadlib' in pkgs
    assert 'msl-network' in pkgs

    # case insensitive
    pkgs = utils._create_install_list(('MSL-loadlib', 'network', 'gtc'), None, None, None, False)
    assert len(pkgs) == 3
    assert 'msl-loadlib' in pkgs
    assert 'msl-network' in pkgs
    assert 'GTC' in pkgs

    # msl-network does not show up twice
    pkgs = utils._create_install_list(('loadlib', 'msl-network', 'network', 'pr-omega-logger'), None, None, None, False)
    assert len(pkgs) == 3
    assert 'msl-loadlib' in pkgs
    assert 'msl-network' in pkgs
    assert 'pr-omega-logger' in pkgs

    # all repos that start with "pr-"
    pkgs = utils._create_install_list(('pr-*',), None, None, None, False)
    assert len(pkgs) >= 1
    assert 'pr-omega-logger' in pkgs
    for p in pkgs:
        assert p.startswith('pr-')

    # all Python repos
    pkgs = utils._create_install_list(('*',), None, None, None, False)
    assert 'msl-loadlib' in pkgs
    assert 'hs-logger' in pkgs
    assert 'pr-omega-logger' in pkgs
    assert 'GTC' in pkgs
    assert 'Lengthstds_Environment_Monitoring' not in pkgs  # it's written in C# not Python

    # the asterisk appears first before any other character
    pkgs = utils._create_install_list(('*loadlib',), None, None, None, False)
    assert len(pkgs) == 1
    assert pkgs['msl-loadlib']['version_requested'] is None
    assert pkgs['msl-loadlib']['extras_require'] is None

    #
    # support for extras_require and version ranges
    #
    names = ('loadlib[clr]', 'gtc', 'Lengthstds_Environment_Monitoring', 'Does Not Exist', 'msl-network==0.4')
    pkgs = utils._create_install_list(names, None, None, None, False)
    assert len(pkgs) == 3
    assert 'msl-loadlib' in pkgs
    assert 'GTC' in pkgs
    assert 'msl-network' in pkgs
    assert 'Lengthstds_Environment_Monitoring' not in pkgs  # it's written in C# not Python
    assert 'Does Not Exist' not in pkgs
    assert pkgs['msl-loadlib']['extras_require'] == '[clr]'
    assert pkgs['msl-loadlib']['version_requested'] is None
    assert pkgs['msl-network']['extras_require'] is None
    assert pkgs['msl-network']['version_requested'] == '==0.4'

    names = ('msl-loadlib[java,clr]>=0.4,<0.6', 'GtC==1.1.*', 'network!=0.4.0', 'ms*qt[pyside]==0.1.0.dev0')
    pkgs = utils._create_install_list(names, None, None, None, False)
    assert len(pkgs) == 4
    assert pkgs['msl-loadlib']['extras_require'] == '[java,clr]'
    assert pkgs['msl-loadlib']['version_requested'] == '>=0.4,<0.6'
    assert pkgs['GTC']['extras_require'] is None
    assert pkgs['GTC']['version_requested'] == '==1.1.*'
    assert pkgs['msl-network']['extras_require'] is None
    assert pkgs['msl-network']['version_requested'] == '!=0.4.0'
    assert pkgs['msl-qt']['extras_require'] == '[pyside]'
    assert pkgs['msl-qt']['version_requested'] == '==0.1.0.dev0'


def test_has_git():
    assert utils.has_git


def test_outdated_pypi_packages():
    outdated = utils.outdated_pypi_packages()
    for info in outdated.values():
        assert info['installed_version']
        assert isinstance(info['using_pypi'], bool) and info['using_pypi']
        assert info['extras_require'] == ''
        assert info['version']
        assert info['repo_name'] == ''



"""
Command line interface for the :ref:`uninstall <uninstall-cli>` command.
"""
from .cli_argparse import add_argument_all
from .cli_argparse import add_argument_disable_mslpm_version_check
from .cli_argparse import add_argument_package_names
from .cli_argparse import add_argument_quiet
from .cli_argparse import add_argument_yes
from .uninstall import uninstall

DESCRIPTION = '{} MSL packages.'

EXAMPLE = """All other optional arguments are passed to "pip uninstall".

Examples:
    msl {0} loadlib
    msl {0} qt --no-python-version-warning
"""


def add_parser_uninstall(parser, name='uninstall'):
    """Add the :ref:`uninstall <uninstall-cli>` command to the parser."""
    if name == 'uninstall':
        p = parser.add_parser(
            name,
            help='Uninstall MSL packages.',
            description=DESCRIPTION.format(name.capitalize()),
            epilog=EXAMPLE.format(name),
        )
    else:
        p = parser.add_parser(
            name,
            help='Alias for uninstall.',
            description=DESCRIPTION.format(name.capitalize()),
            epilog=EXAMPLE.format(name),
        )
    add_argument_package_names(p)
    add_argument_all(p)
    add_argument_yes(p)
    add_argument_quiet(p)
    add_argument_disable_mslpm_version_check(p)
    p.set_defaults(func=execute)


def execute(args, parser):
    """Executes the :ref:`uninstall <uninstall-cli>` command."""
    if parser.contains_package_names():
        uninstall(*args.names, yes=args.yes, pip_options=args.pip_options)



"""
Custom argument parsers.
"""
import argparse

from .utils import log


class ArgumentParser(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        """A custom argument parser."""
        kwargs['add_help'] = False  # use a custom help message (see below)
        kwargs['formatter_class'] = argparse.RawTextHelpFormatter
        super(ArgumentParser, self).__init__(*args, **kwargs)

        # so that capitalization and periods are consistent in the help
        self.add_argument(
            '-h', '--help',
            action='help',
            help='Show this help message and exit.',
            default=argparse.SUPPRESS
        )

    def get_command_name(self):
        """:class:`str`: Returns the name of the command, e.g., ``install``, ``list``, ..."""
        return self.prog.split()[-1]

    def contains_package_names(self, quiet=False):
        """Check whether package names were specified or the ``--all`` flag was used.

        .. versionchanged:: 2.5.0
           Added the `quiet` keyword argument.

        Parameters
        ----------
        quiet : :class:`bool`
            Whether to suppress the error message from being shown.

        Returns
        -------
        :class:`bool`
            Whether package names were specified or the ``--all`` flag was used.
        """
        args = self.parse_known_args()[0]
        if not args.all and not args.names:
            if not quiet:
                non_msl_flag = ''
                if args.cmd in ['update', 'upgrade']:
                    non_msl_flag = ' and/or the --non-msl flag'
                log.error('You must specify the MSL package name(s) to %s or use '
                          'the --all flag%s', args.cmd, non_msl_flag)
            return False
        return True


def add_argument_all(parser):
    """Add an ``--all`` argument to the parser."""
    parser.add_argument(
        '-a', '--all',
        action='store_true',
        default=False,
        help='{} all MSL packages.'.format(parser.get_command_name().capitalize()),
    )


def add_argument_branch(parser):
    """Add a ``--branch`` argument to the parser."""
    parser.add_argument(
        '-b', '--branch',
        help='The git branch to use to {} the package(s).'.format(parser.get_command_name()),
    )


def add_argument_package_names(parser):
    """Add a ``--names`` argument to the parser."""
    parser.add_argument(
        'names',
        nargs='*',
        help='The name(s) of the MSL package(s) to {}.\n'
             'The "msl-" prefix can be omitted (e.g., loadlib is\n'
             'equivalent to msl-loadlib). Also accepts shell-style\n'
             'wildcards (e.g., pr-*).'.format(parser.get_command_name())
    )


def add_argument_quiet(parser):
    """Add a ``--quiet`` argument to the parser."""
    parser.add_argument(
        '-q', '--quiet',
        action='count',
        default=0,
        help='Give less output. Option is additive, and can\n'
             'be used up to 4 times (which corresponds to\n'
             'silencing DEBUG, INFO, WARNING and ERROR\n'
             'logging levels).'
    )


def add_argument_tag(parser):
    """Add a ``--tag`` argument to the parser."""
    parser.add_argument(
        '-t', '--tag',
        help='The git tag to use to {} the package(s).'.format(parser.get_command_name()),
    )


def add_argument_update_cache(parser):
    """Add an ``--update-cache`` argument to the parser."""
    parser.add_argument(
        '-u', '--update-cache',
        action='store_true',
        default=False,
        help='Force the GitHub and PyPI caches to be updated.\n'
             'The information about MSL packages are cached\n'
             'for subsequent calls to the {} command. After 24\n'
             'hours the cache is automatically updated. Include\n'
             'this flag to force the cache to be updated now.'
             .format(parser.get_command_name()),
    )


def add_argument_yes(parser):
    """Add a ``--yes`` argument to the parser."""
    parser.add_argument(
        '-y', '--yes',
        action='store_true',
        default=False,
        help='Don\'t ask for confirmation to {} the package(s).'
             .format(parser.get_command_name()),
    )


def add_argument_disable_mslpm_version_check(parser):
    """Add a ``--disable-mslpm-version-check`` argument to the parser."""
    parser.add_argument(
        '-D', '--disable-mslpm-version-check',
        action='store_true',
        default=False,
        help='Don\'t check if there is a new version of the\n'
             'MSL-Package-Manager available.',
    )


def add_argument_commit(parser):
    """Add a ``--commit`` argument to the parser."""
    parser.add_argument(
        '-c', '--commit',
        help='The hash value of a git commit to use to {}\n'
             'a package.'.format(parser.get_command_name()),
    )



import msl.package_manager as pm


def test_pypi():
    pkgs = pm.pypi()
    assert len(pkgs) == 6
    assert pm.utils._PKG_NAME in pkgs
    assert 'msl-io' in pkgs
    assert 'msl-loadlib' in pkgs
    assert 'msl-network' in pkgs
    assert 'GTC' in pkgs
    assert 'Quantity-Value' in pkgs



"""
Create a new package.
"""
import os
import time

from . import utils


def create(*names, **kwargs):
    """Create a new package.

    Parameters
    ----------
    *names
        The name(s) of the package(s) to create.
    **kwargs
        * author -- :class:`str`
            The name of the author to use for the new package. If :data:`None`
            then uses :func:`.utils.get_username` to determine the author's name.
            Default is :data:`None`.
        * directory -- :class:`str`
            The directory to create the new package(s) in. If :data:`None`
            then creates the new package(s) in the current working directory.
            Default is :data:`None`.
        * email -- :class:`str`
            The author's email address. If :data:`None` then uses
            :func:`.utils.get_email` to determine the author's email address.
            Default is :data:`None`.
        * namespace -- :class:`str`
            The namespace that the package belongs to. If `namespace` is :data:`None`
            or an empty string then create a new package that is not part of a namespace.
            Default is the ``'msl'`` namespace.
        * yes -- :class:`bool`
            If :data:`True` then don't ask for verification for the `author` name
            and for the `email` address. This argument is only used if you do not
            specify the `author` or the `email` value. The verification step allows
            you to change the value that was automatically determined before the
            package is created. The default is to ask for verification before creating
            the package if the `author` or the `email` value was not specified.
            Default is :data:`False`.

    """
    # TODO Python 2.7 does not support named arguments after using *args
    #  we can define yes=False, author=None, email=None, directory=None in the
    #  function signature if we choose to drop support for Python 2.7
    utils._check_kwargs(kwargs, {'yes', 'author', 'email', 'directory', 'namespace'})

    yes = kwargs.get('yes', False)
    author = kwargs.get('author', None)
    email = kwargs.get('email', None)
    directory = kwargs.get('directory', None)
    namespace = kwargs.get('namespace', 'msl')
    if namespace:
        namespace = namespace.lower().replace('-', '')
    else:
        namespace = ''

    # ensure that the names contain valid characters for a python package
    # and that the folder does not already exist in the current working directory
    _names = [n.replace(' ', '_') for n in names]

    if directory is None:
        directory = os.getcwd()

    roots, pkg_names = [], []
    for name in _names:
        if namespace and name.lower().startswith(namespace + '-'):
            name = name[len(namespace)+1:]

        if name[0].isdigit():
            utils.log.warning('A package name cannot start with a number: ignored %r', name)
            continue

        keep = True
        for c in name:
            if not (c.isalnum() or c == '_' or c == '-'):
                utils.log.warning('A package name can only contain letters, numbers and underscores: '
                                  'ignored %r', name)
                keep = False
                break

        if keep:
            if namespace:
                basename = namespace + '-' + name
            else:
                basename = name
            root = os.path.join(directory, basename)
            if os.path.isdir(root):
                utils.log.warning('A %r folder already exists: ignored %r', root, name)
            else:
                roots.append(root)
                pkg_names.append(name)

    if not pkg_names:
        return

    # determine the author's name
    if author is not None:
        if isinstance(author, str):
            author_name = author
        elif isinstance(author, (list, tuple)):
            author_name = ' '.join(author)
        else:
            raise TypeError('The author\'s name must be either a string or a list of strings')
    else:
        author_name = utils.get_username()
        if not yes:
            try:
                new_name = utils._get_input('You can enter a new author name [default: {}]: '.format(author_name))
                if new_name:
                    author_name = new_name
            except:
                utils.log.error('Aborted -- cannot create MSL package')
                return

    # determine the author's email address
    if email is not None:
        if isinstance(email, str):
            email_address = email
        else:
            raise TypeError('The email address must be a string')
    else:
        email_address = utils.get_email()
        if not yes:
            try:
                new_email = utils._get_input('You can enter a new email address [default: {}]: '.format(email_address))
                if new_email:
                    email_address = new_email
            except:
                utils.log.error('Aborted -- cannot create MSL package')
                return

    # create the new package
    template_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), 'template'))
    for msl_root, package in zip(roots, pkg_names):

        safe_pkg = package.replace('-', '_')

        aliases = {
            '${namespace}': namespace,
            '${package}': safe_pkg,
            '${namespace-or-package}': namespace if namespace else safe_pkg,
            '${full-name}': namespace + '-' + package if namespace else package,
            '${author}': author_name,
            '${year}': time.strftime('%Y'),
            '${email}': str(email_address),
            '${=}': '='*(len(os.path.basename(msl_root))),  # used to underline the header in a *.rst file
            '${sphinx-import}': 'from ' + namespace + ' import ' + safe_pkg if namespace else 'import ' + safe_pkg,
            '${installed-name}': namespace + '.' + safe_pkg if namespace else safe_pkg,
        }

        for root, dirs, files in os.walk(template_dir):
            if not namespace and '$' in root:
                # will create the source folder below
                continue

            new_dir = msl_root + root.split(template_dir)[1]
            new_dir = new_dir.replace('${namespace}', aliases['${namespace}'])
            new_dir = new_dir.replace('${package}', aliases['${package}'])
            if not os.path.isdir(new_dir):
                os.makedirs(new_dir)

            for filename in files:
                with open(os.path.join(root, filename), mode='rt') as fp:
                    lines = fp.read()

                if not namespace:
                    if filename == '.coveragerc':
                        lines = lines.replace('omit =\n', '')
                        lines = lines.replace('${namespace}/examples/*\n', '')
                    elif filename == 'setup.cfg':
                        lines = lines.replace('--ignore ${namespace}/examples\n', '')
                    elif filename == 'setup.py.template':
                        lines = lines.replace('${namespace}/${package}', safe_pkg)
                        lines = lines.replace('${namespace}*', safe_pkg)

                for alias in aliases:
                    lines = lines.replace(alias, aliases[alias])

                with open(os.path.join(new_dir, filename.replace('.template', '')), mode='wt') as fp:
                    fp.write(lines)

        if not namespace:
            # create the source folder
            src_dir = os.path.join(msl_root, package.replace('-', '_'))
            os.makedirs(src_dir)
            init = os.path.join(template_dir, '${namespace}', '${package}', '__init__.py.template')
            with open(init, mode='rt') as fp:
                lines = fp.read()
            for alias in aliases:
                lines = lines.replace(alias, aliases[alias])
            with open(os.path.join(src_dir, '__init__.py'), mode='wt') as fp:
                fp.write(lines)

        fullname = '{}-{}'.format(namespace, package) if namespace else package
        if os.path.isdir(msl_root):
            utils.log.info('Created %s in %r', fullname, msl_root)
        else:
            utils.log.error('Error creating %r', fullname)



import msl.package_manager as pm


def test_github():
    pkgs = pm.github()
    assert pm.utils._PKG_NAME in pkgs
    pkg = pkgs[pm.utils._PKG_NAME]
    assert len(pkg['description']) > 0
    assert len(pkg['version']) > 0
    assert len(pkg['tags']) > 10
    assert len(pkg['branches']) > 0



import importlib
import re
import sys
try:
    from importlib import reload
except ImportError:  # then Python 2
    from imp import reload

import pytest

from msl.package_manager import install
from msl.package_manager import installed
from msl.package_manager import uninstall


def cleanup():
    if 'msl-loadlib' in installed():
        uninstall('msl-loadlib', yes=True)


def teardown_module():
    cleanup()


@pytest.mark.skipif(
    sys.version_info[:2] == (2, 7) and sys.platform == 'darwin',
    reason='hyphen gets appended to msl-loadlib version (0.7.0-)'
)
def test_version_suffix():
    cleanup()

    install('loadlib==0.7.0', yes=True)

    # reload all msl modules
    for name, module in sys.modules.copy().items():
        if name.startswith('msl'):
            try:
                reload(module)
            except:
                pass

    packages = installed()

    # the version does not end with a commit hash
    assert 'msl-loadlib' in packages
    assert packages['msl-loadlib']['version'] == '0.7.0'
    msl_loadlib = importlib.import_module('msl.loadlib')
    assert msl_loadlib.__version__ == '0.7.0'

    # msl-package-manager gets installed in editable mode (pip install -e)
    # in cloud-based testing platforms
    if 'msl-package-manager' in packages:
        version = packages['msl-package-manager']['version']
        if 'dev' in version:
            assert version.endswith('+editable')
            assert version.count('+') == 1

    uninstall('loadlib', yes=True)
    assert 'msl-loadlib' not in installed()

    install('loadlib', branch='main', yes=True)
    packages = installed()

    # the version ends with a commit hash
    assert 'msl-loadlib' in packages
    version = packages['msl-loadlib']['version']
    assert re.search(r'\+[a-z0-9]{7}$', version)
    assert version.count('+') == 1
    assert reload(msl_loadlib).__version__ == version

    uninstall('loadlib', yes=True)
    assert 'msl-loadlib' not in installed()



import logging

from msl.package_manager import cli
from msl.package_manager import utils


def get_args(command):
    return cli.parse_args(command.split())


def test_install_update_args():

    for cmd in ('install', 'update', 'upgrade'):
        cmd += ' '

        args = get_args(cmd)
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-u')
        assert not args.names
        assert not args.all
        assert args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--update-cache')
        assert not args.names
        assert not args.all
        assert args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-b thebranch')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch == 'thebranch'
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--branch thebranch --yes')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch == 'thebranch'
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-t thetag')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag == 'thetag'
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--tag thetag -qy')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag == 'thetag'
        assert args.quiet == 1
        assert args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-c 123456789')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit == '123456789'
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--commit 123456789')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit == '123456789'
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-t thetag --branch thebranch -c abcd')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch == 'thebranch'
        assert args.commit == 'abcd'
        assert args.tag == 'thetag'
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-a')
        assert not args.names
        assert args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--all')
        assert not args.names
        assert args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + 'package')
        assert len(args.names) == 1 and args.names[0] == 'package'
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--all -t thetag --branch thebranch a bb cdfg package --update-cache')
        assert len(args.names) == 4
        assert args.names[0] == 'a'
        assert args.names[1] == 'bb'
        assert args.names[2] == 'cdfg'
        assert args.names[3] == 'package'
        assert args.all
        assert args.update_cache
        assert args.branch == 'thebranch'
        assert args.commit is None
        assert args.tag == 'thetag'
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--log here.txt -t thetag --no-color --branch thebranch '
                              '--index-url https://my.index.org/simple -c abc --user')
        assert not args.names
        assert not args.all
        assert not args.update_cache
        assert args.branch == 'thebranch'
        assert args.commit == 'abc'
        assert args.tag == 'thetag'
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 6
        assert args.pip_options[0] == '--log'
        assert args.pip_options[1] == 'here.txt'
        assert args.pip_options[2] == '--no-color'
        assert args.pip_options[3] == '--index-url'
        assert args.pip_options[4] == 'https://my.index.org/simple'
        assert args.pip_options[5] == '--user'

        args = get_args(cmd + 'io --log here.txt equipment -t thetag --no-color network --constraint file.txt '
                              '--branch thebranch -i https://my.index.org/simple --user loadlib --commit qt')
        assert args.names == ['io', 'equipment', 'network', 'loadlib']
        assert not args.all
        assert not args.update_cache
        assert args.branch == 'thebranch'
        assert args.commit == 'qt'
        assert args.tag == 'thetag'
        assert args.quiet == 0
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 8
        assert args.pip_options[0] == '--log'
        assert args.pip_options[1] == 'here.txt'
        assert args.pip_options[2] == '--no-color'
        assert args.pip_options[3] == '--constraint'
        assert args.pip_options[4] == 'file.txt'
        assert args.pip_options[5] == '-i'
        assert args.pip_options[6] == 'https://my.index.org/simple'
        assert args.pip_options[7] == '--user'

        args = get_args(cmd + 'io network --compile -qq --upgrade-strategy eager')
        assert args.names == ['io', 'network']
        assert not args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 2
        assert not args.yes
        assert not args.disable_mslpm_version_check
        assert args.pip_options == ['--compile', '--upgrade-strategy', 'eager']

        args = get_args(cmd + '--no-deps -f file://home/ --all -y --retries 10')
        assert not args.names
        assert args.all
        assert not args.update_cache
        assert args.branch is None
        assert args.commit is None
        assert args.tag is None
        assert args.quiet == 0
        assert args.yes
        assert not args.disable_mslpm_version_check
        assert args.pip_options == ['--no-deps', '-f', 'file://home/', '--retries', '10']


def test_uninstall_args():

    for cmd in ('uninstall', 'remove'):
        cmd += ' '

        args = get_args(cmd)
        assert not args.all
        assert not args.names
        assert not args.yes
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '-a')
        assert args.all
        assert not args.names
        assert not args.yes
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--all --yes')
        assert args.all
        assert not args.names
        assert args.yes
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 0

        args = get_args(cmd + 'package1 --log here.txt package2')
        assert not args.all
        assert args.names[0] == 'package1'
        assert args.names[1] == 'package2'
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert not args.yes
        assert len(args.pip_options) == 2
        assert args.pip_options[0] == '--log'
        assert args.pip_options[1] == 'here.txt'

        args = get_args(cmd + '-a packageA -y')
        assert args.all
        assert args.names[0] == 'packageA'
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert args.yes
        assert len(args.pip_options) == 0

        args = get_args(cmd + '--no-color -a packageA -y')
        assert args.all
        assert args.names[0] == 'packageA'
        assert args.yes
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 1
        assert args.pip_options[0] == '--no-color'

        args = get_args(cmd + 'qt -r reqs.txt io --no-color loadlib')
        assert not args.all
        assert args.names == ['qt', 'io', 'loadlib']
        assert not args.yes
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 3
        assert args.pip_options[0] == '-r'
        assert args.pip_options[1] == 'reqs.txt'
        assert args.pip_options[2] == '--no-color'


def test_list_args():

    args = get_args('list')
    assert not args.github
    assert not args.json
    assert not args.pypi
    assert not args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list -g')
    assert args.github
    assert not args.json
    assert not args.pypi
    assert not args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list --github')
    assert args.github
    assert not args.json
    assert not args.pypi
    assert not args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list --update-cache')
    assert not args.github
    assert not args.json
    assert not args.pypi
    assert args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list -u')
    assert not args.github
    assert not args.json
    assert not args.pypi
    assert args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list -j')
    assert not args.github
    assert args.json
    assert not args.pypi
    assert not args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list --json')
    assert not args.github
    assert args.json
    assert not args.pypi
    assert not args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list -j -g --update-cache')
    assert args.github
    assert args.json
    assert not args.pypi
    assert args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list -jgu')
    assert args.github
    assert args.json
    assert not args.pypi
    assert args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list -p --update-cache')
    assert not args.github
    assert not args.json
    assert args.pypi
    assert args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list --pypi --update-cache')
    assert not args.github
    assert not args.json
    assert args.pypi
    assert args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('list --invalid-pip-option --does-not-get-parsed-by-pip')
    assert not args.github
    assert not args.json
    assert not args.pypi
    assert not args.update_cache
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 2
    assert args.pip_options[0] == '--invalid-pip-option'
    assert args.pip_options[1] == '--does-not-get-parsed-by-pip'


def test_create_args():

    args = get_args('create')
    assert not args.names
    assert args.author is None
    assert args.email is None
    assert args.dir is None
    assert args.namespace is None
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('create -a first m. last -e my@email.com')
    assert not args.names
    assert len(args.author) == 3
    assert args.author[0] == 'first'
    assert args.author[1] == 'm.'
    assert args.author[2] == 'last'
    assert args.email == 'my@email.com'
    assert args.dir is None
    assert args.namespace is None
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('create --dir D:\\')
    assert not args.names
    assert args.author is None
    assert args.email is None
    assert args.dir == 'D:\\'
    assert args.namespace is None
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('create -a first m. last -e my@email.com --dir the/folder p1 pkg2 package3 d')
    assert len(args.names) == 4
    assert args.names[0] == 'p1'
    assert args.names[1] == 'pkg2'
    assert args.names[2] == 'package3'
    assert args.names[3] == 'd'
    assert len(args.author) == 3
    assert args.author[0] == 'first'
    assert args.author[1] == 'm.'
    assert args.author[2] == 'last'
    assert args.email == 'my@email.com'
    assert args.dir == 'the/folder'
    assert args.namespace is None
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('create --namespace pr')
    assert not args.names
    assert args.author is None
    assert args.email is None
    assert args.dir is None
    assert args.namespace == 'pr'
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('create -n pr')
    assert not args.names
    assert args.author is None
    assert args.email is None
    assert args.dir is None
    assert args.namespace == 'pr'
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 0

    args = get_args('create --invalid-pip-option XXX --does-not-get-parsed-by-pip')
    assert args.names == ['XXX']
    assert args.author is None
    assert args.email is None
    assert args.dir is None
    assert args.namespace is None
    assert args.quiet == 0
    assert not args.disable_mslpm_version_check
    assert len(args.pip_options) == 2
    assert args.pip_options[0] == '--invalid-pip-option'
    assert args.pip_options[1] == '--does-not-get-parsed-by-pip'


def test_quiet():

    args = get_args('list')
    assert args.quiet == 0
    assert utils.log.level == logging.DEBUG
    assert utils._pip_quiet == 0

    args = get_args('list --quiet')
    assert args.quiet == 1
    assert utils.log.level == logging.INFO
    assert utils._pip_quiet == 0

    args = get_args('list -q')
    assert args.quiet == 1
    assert utils.log.level == logging.INFO
    assert utils._pip_quiet == 0

    args = get_args('list --quiet --quiet')
    assert args.quiet == 2
    assert utils.log.level == logging.WARNING
    assert utils._pip_quiet == 1

    args = get_args('list -q -q')
    assert args.quiet == 2
    assert utils.log.level == logging.WARNING
    assert utils._pip_quiet == 1

    args = get_args('list -qq')
    assert args.quiet == 2
    assert utils.log.level == logging.WARNING
    assert utils._pip_quiet == 1

    args = get_args('list --quiet --quiet --quiet')
    assert args.quiet == 3
    assert utils.log.level == logging.ERROR
    assert utils._pip_quiet == 2

    args = get_args('list -q -q -q')
    assert args.quiet == 3
    assert utils.log.level == logging.ERROR
    assert utils._pip_quiet == 2

    args = get_args('list -qqq')
    assert args.quiet == 3
    assert utils.log.level == logging.ERROR
    assert utils._pip_quiet == 2

    args = get_args('list --quiet --quiet --quiet --quiet')
    assert args.quiet == 4
    assert utils.log.level > logging.CRITICAL
    assert utils._pip_quiet == 3

    args = get_args('list -q -q -q -q')
    assert args.quiet == 4
    assert utils.log.level > logging.CRITICAL
    assert utils._pip_quiet == 3

    args = get_args('list -qqqq')
    assert args.quiet == 4
    assert utils.log.level > logging.CRITICAL
    assert utils._pip_quiet == 3

    args = get_args('list --quiet --quiet --quiet --quiet --quiet')
    assert args.quiet == 5
    assert utils.log.level > logging.CRITICAL
    assert utils._pip_quiet == 3  # 3 is the maximum

    args = get_args('list -q -q -q -q -q')
    assert args.quiet == 5
    assert utils.log.level > logging.CRITICAL
    assert utils._pip_quiet == 3  # 3 is the maximum

    args = get_args('list -qqqqq')
    assert args.quiet == 5
    assert utils.log.level > logging.CRITICAL
    assert utils._pip_quiet == 3  # 3 is the maximum


def test_authorise():
    for command in ['authorise', 'authorize']:
        args = get_args(command + ' --invalid-pip-option notused --does-not-get-parsed-by-pip')
        assert args.quiet == 0
        assert not args.disable_mslpm_version_check
        assert len(args.pip_options) == 3
        assert args.pip_options[0] == '--invalid-pip-option'
        assert args.pip_options[1] == 'notused'
        assert args.pip_options[2] == '--does-not-get-parsed-by-pip'



"""
Install, uninstall, update, list and create MSL packages.

The following constants are available.
"""
import re
from collections import namedtuple

from .authorise import authorise
from .create import create
from .install import install
from .uninstall import uninstall
from .update import update
from .utils import github
from .utils import info
from .utils import installed
from .utils import outdated_pypi_packages
from .utils import pypi
from .utils import set_log_level

__author__ = 'Measurement Standards Laboratory of New Zealand'
__copyright__ = '\xa9 2017 - 2023, ' + __author__
__version__ = '2.5.3'

_v = re.search(r'(\d+)\.(\d+)\.(\d+)[.-]?(.*)', __version__).groups()

version_info = namedtuple('version_info', 'major minor micro releaselevel')(int(_v[0]), int(_v[1]), int(_v[2]), 'final')
""":obj:`~collections.namedtuple`: Contains the version information as a (major, minor, micro, releaselevel) tuple."""



"""
Command line interface for the :ref:`authorise <authorise-cli>` command.
"""
from .authorise import WARNING_MESSAGE
from .authorise import authorise
from .cli_argparse import add_argument_disable_mslpm_version_check
from .cli_argparse import add_argument_quiet

HELP = 'Enable authorisation to the GitHub API.'

DESCRIPTION = HELP + """

When requesting information about the MSL repositories that are
available on GitHub there is a limit to how often you can send
requests to the GitHub API. If you have a GitHub account and 
include your username and a personal access token with each 
request then this limit is increased.

Running this command will create a file that contains your GitHub
username and a personal access token so that requests to the GitHub
API are authorised.

***************************** IMPORTANT ******************************
{}
**********************************************************************
""".format(WARNING_MESSAGE)

EXAMPLE = """
Example:

  $ msl {}
"""


def add_parser_authorise(parser, name='authorise'):
    """Add the :ref:`authorise <authorise-cli>` command to the parser."""
    p = parser.add_parser(
        name,
        help=HELP if name == 'authorise' else 'Alias for authorise.',
        description=DESCRIPTION,
        epilog=EXAMPLE.format(name),
    )
    add_argument_quiet(p)
    add_argument_disable_mslpm_version_check(p)
    p.set_defaults(func=execute)


def execute(args, parser):
    """Executes the :ref:`authorise <authorise-cli>` command."""
    print('For more detailed instructions on how to generate a personal access token see')
    print('https://docs.github.com/en/github/authenticating-to-github/creating-a-personal-access-token')
    print()
    print('To generate a personal access token, perform the following steps:')
    print('  1. Login to your GitHub account')
    print('  2. Go to Settings -> Developer settings -> Personal access tokens')
    print('  3. Click "Generate new token"')
    print('  4. In the "Note" field enter a description for the token (e.g., msl-package-manager)')
    print('  5. Select the expiration date of your token')
    print('  6. Optional: Select the scopes that you want to associate with the token')
    print('  7. Click "Generate token"')
    print('  8. Copy the token to your clipboard and paste it in the terminal when requested')
    print()
    authorise()



import json
import sys

import colorama
import pytest

from msl.package_manager import install
from msl.package_manager import installed
from msl.package_manager import uninstall
from msl.package_manager import update
from msl.package_manager import utils


@pytest.mark.skipif(sys.version_info[:2] < (3, 6), reason='only test in Python 3.6+')
def test_issue8_a():
    repo = 'pr-omega-logger'
    pkg = 'omega-logger'
    assert pkg not in installed()
    install(repo, yes=True)
    assert pkg in installed()
    update(pkg, branch='main', yes=True)
    assert pkg in installed()

    # msl-equipment, msl-io and msl-loadlib are also installed
    # do not include any packages in the call to uninstall() to
    # uninstall all MSL packages that were installed in this test
    uninstall(yes=True)
    installed_ = installed()
    assert pkg not in installed_
    assert 'msl-equipment' not in installed_
    assert 'msl-loadlib' not in installed_
    assert 'msl-io' not in installed_


@pytest.mark.skipif(sys.version_info[:2] != (2, 7), reason='only test in Python 2.7')
def test_issue8_b():
    # test issue 8 in Python 2.7 since pr-omega-logger requires Python 3.6+
    repo = 'pr-webpage-text'
    pkg = 'webpage-text'
    assert pkg not in installed()
    install(repo, yes=True)
    assert pkg in installed()
    update(pkg, branch='main', yes=True)
    assert pkg in installed()
    uninstall(pkg, yes=True)
    assert pkg not in installed()


def test_issue_11(caplog):
    # installed packages
    utils.info(as_json=True)
    assert len(caplog.messages) == 3
    record = caplog.records[0]
    assert record.levelname == 'DEBUG'
    assert record.message.startswith('Getting the packages from ')
    record = caplog.records[1]
    assert record.levelname == 'DEBUG'
    assert record.message == colorama.Fore.RESET
    record = caplog.records[2]
    assert record.levelname == 'INFO'
    pkgs = json.loads(record.message)
    requires = sorted(pkgs[utils._PKG_NAME]['requires'])
    assert requires == ['colorama', 'setuptools']

    caplog.clear()

    # from PyPI
    utils.info(as_json=True, from_pypi=True)
    assert len(caplog.messages) == 3
    record = caplog.records[0]
    assert record.levelname == 'DEBUG'
    assert record.message == 'Loaded the cached information about the PyPI packages'
    record = caplog.records[1]
    assert record.levelname == 'DEBUG'
    assert record.message == colorama.Fore.RESET
    record = caplog.records[2]
    assert record.levelname == 'INFO'
    pkgs = json.loads(record.message)
    assert utils._PKG_NAME in pkgs
    assert 'requires' not in pkgs[utils._PKG_NAME]
    assert 'version' in pkgs[utils._PKG_NAME]
    assert 'description' in pkgs[utils._PKG_NAME]

    caplog.clear()

    # from GitHub
    utils.info(as_json=True, from_github=True)
    assert len(caplog.messages) == 3
    record = caplog.records[0]
    assert record.levelname == 'DEBUG'
    assert record.message == 'Loaded the cached information about the GitHub repositories'
    record = caplog.records[1]
    assert record.levelname == 'DEBUG'
    assert record.message == colorama.Fore.RESET
    record = caplog.records[2]
    assert record.levelname == 'INFO'
    pkgs = json.loads(record.message)
    assert utils._PKG_NAME in pkgs
    assert 'requires' not in pkgs[utils._PKG_NAME]
    assert 'version' in pkgs[utils._PKG_NAME]
    assert 'description' in pkgs[utils._PKG_NAME]
    assert 'branches' in pkgs[utils._PKG_NAME]
    assert 'tags' in pkgs[utils._PKG_NAME]



import importlib
import sys
try:
    from importlib import reload
except ImportError:  # then Python 2
    from imp import reload

import pytest

from msl.package_manager import install
from msl.package_manager import installed
from msl.package_manager import uninstall
from msl.package_manager import update
from msl.package_manager import utils


def cleanup():
    msl_installed = installed()
    if 'msl-loadlib' in msl_installed:
        uninstall('loadlib', yes=True)
    if 'msl-io' in msl_installed:
        uninstall('io', yes=True)


def teardown_module():
    cleanup()


@pytest.mark.skipif(
    sys.version_info[:2] == (2, 7) and sys.platform == 'darwin',
    reason='hyphen gets appended to msl-loadlib version (0.7.0-)'
)
def test_msl_and_non_msl():
    cleanup()

    # msl-io has xlrd<2.0 as a dependency in commit 29d48bb (afterwards it is bundled)
    # make sure that xlrd is not >2.0 after updating

    install('loadlib==0.7.0', yes=True)
    install('io', commit='29d48bb40ad12b6d2270ac2a69a73293728c089e', yes=True)

    # reload all msl modules
    for name, module in sys.modules.copy().items():
        if name.startswith('msl'):
            try:
                reload(module)
            except:
                pass

    xlrd = importlib.import_module('xlrd')
    assert xlrd.__version__ == '1.2.0'

    outdated = utils.outdated_pypi_packages()
    assert 'msl-loadlib' not in outdated
    assert 'msl-io' not in outdated

    # strange, this assert fails in Python 3.5 and does not
    # depend on the version of pip (checked back to pip 10.0)
    if sys.version_info[:2] != (3, 5):
        assert outdated['xlrd']['version'] == '<2.0'

    msl_loadlib = importlib.import_module('msl.loadlib')
    assert installed()['msl-loadlib']['version'] == '0.7.0'
    assert msl_loadlib.__version__ == '0.7.0'

    update('loadlib==0.8.0', include_non_msl=True, yes=True)

    assert reload(xlrd).__version__ == '1.2.0'

    assert installed()['msl-loadlib']['version'] == '0.8.0'
    assert reload(msl_loadlib).__version__ == '0.8.0'

    uninstall('loadlib', 'io', yes=True)
    installed_ = installed()
    assert 'msl-loadlib' not in installed_
    assert 'msl-io' not in installed_


def test_from_commit():
    cleanup()

    commit1 = 'b2581039e4e3f2ffe2599927d589809efbb9a1ff'
    commit2 = '12591bade80321c3a165f7a7364ef13f568d622b'

    install('loadlib', yes=True, commit=commit1)

    # reload all msl modules
    for name, module in sys.modules.copy().items():
        if name.startswith('msl'):
            try:
                reload(module)
            except:
                pass

    # the version ends with the expected commit hash
    packages = installed()
    expected = '0.9.0.dev0+{}'.format(commit1[:7])
    assert 'msl-loadlib' in packages
    assert packages['msl-loadlib']['version'] == expected
    msl_loadlib = importlib.import_module('msl.loadlib')
    assert msl_loadlib.__version__ == expected

    update('loadlib', yes=True, commit=commit2)

    # reload all msl modules
    for name, module in sys.modules.copy().items():
        if name.startswith('msl'):
            try:
                reload(module)
            except:
                pass

    # the version ends with the expected commit hash
    packages = installed()
    expected = '0.9.1.dev0+{}'.format(commit2[:7])
    assert 'msl-loadlib' in packages
    assert packages['msl-loadlib']['version'] == expected
    msl_loadlib = importlib.import_module('msl.loadlib')
    assert msl_loadlib.__version__ == expected

    uninstall('loadlib', yes=True)
    assert 'msl-loadlib' not in installed()
